from datetime import datetime
from traceback import print_exc

import pandas as pd
from pandas import DataFrame

from offline_results.common.constants import INNER, CREATED_ON, PAY_TV_CONTENT, NO_PAY_TV_CONTENT, CONTENT_ID, STATUS, \
    HAS_HOMEPAGE, HOMEPAGE_ID, HOMEPAGE_STATUS, HOMEPAGE, IS_CONNECTED, ACTIVE, CLUSTER_ID, RECORDS, PAY_TV, REC_TYPE
from offline_results.utils.logger import Logging


class RecencyUtils:

    @staticmethod
    def get_inner_merged_df(data1, data2, on):
        try:
            df = pd.merge(data1, data2, on=on, how=INNER)
            return df
        except Exception as e:
            Logging.error(f"Error while merging on {on}, Error: {e}")

    @staticmethod
    def add_created_on_attribute(dataframe) -> DataFrame:
        try:
            dataframe[CREATED_ON] = datetime.utcnow().isoformat()
            return dataframe
        except Exception as e:
            Logging.error(f"Error while adding time created attribute, Error: {e}")

    @staticmethod
    def add_recommendation_type_attribute(dataframe, rec_type) -> DataFrame:
        try:
            dataframe[REC_TYPE] = rec_type
            return dataframe
        except Exception as e:
            Logging.error(f"Error while adding recommendation type attribute, Error: {e}")

    @staticmethod
    def get_label_wise_homepage_for_contents(
            graph, user_label
    ) -> DataFrame:
        try:
            content_label = PAY_TV_CONTENT if user_label == PAY_TV else NO_PAY_TV_CONTENT
            content_response = graph.custom_query(
                query=f'''g.V().hasLabel('{content_label}').match(
                __.as("c").valueMap('{CONTENT_ID}','{STATUS}').as('{CONTENT_ID}'),
                __.as("c").out('{HAS_HOMEPAGE}')
                .valueMap('{HOMEPAGE_ID}','{HOMEPAGE_STATUS}').as('{HOMEPAGE_ID}')
                ).select('{CONTENT_ID}', '{HOMEPAGE_ID}')''',
                payload={
                    content_label: content_label,
                    CONTENT_ID: CONTENT_ID,
                    HAS_HOMEPAGE: HAS_HOMEPAGE,
                    HOMEPAGE_ID: HOMEPAGE_ID,
                    HOMEPAGE_STATUS: HOMEPAGE_STATUS,
                    STATUS: STATUS
                })
            content_homepage_map_df = pd.DataFrame()
            for query in content_response:
                for data in query:
                    content_id = data[CONTENT_ID][CONTENT_ID][0]
                    content_status = data[CONTENT_ID][STATUS][0]
                    homepage_id = data[HOMEPAGE_ID][HOMEPAGE_ID][0]
                    homepage_status = data[HOMEPAGE_ID][HOMEPAGE_STATUS][0]
                    content_homepage_map = pd.DataFrame([{CONTENT_ID: content_id,
                                                          STATUS: content_status,
                                                          HOMEPAGE_ID: homepage_id,
                                                          HOMEPAGE_STATUS: homepage_status}])
                    content_homepage_map_df = pd.concat([content_homepage_map_df, content_homepage_map], axis=0). \
                        reset_index(drop=True)
            content_homepage_map_df = (
                content_homepage_map_df[[CONTENT_ID, HOMEPAGE_ID]].sort_values([CONTENT_ID, HOMEPAGE_ID])) \
                .reset_index(drop=True)
        except Exception:
            print_exc()
            content_homepage_map_df = []
        return content_homepage_map_df

    @staticmethod
    def get_static_homepage_data(graph):
        try:
            response = graph.custom_query(
                query=f'''g.V().hasLabel('{HOMEPAGE}').match(
                                    __.as("c").values('{CREATED_ON}').as('{CREATED_ON}'),
                                    __.as("c").values('{HOMEPAGE_ID}').as('{HOMEPAGE_ID}'),
                                    __.as("c").values('{HOMEPAGE_STATUS}').as('{HOMEPAGE_STATUS}'),
                                    __.as("c").values('{IS_CONNECTED}').as('{IS_CONNECTED}')
                                    ).select('{CREATED_ON}','{HOMEPAGE_ID}','{HOMEPAGE_STATUS}','{IS_CONNECTED}')''',
                payload={
                    HOMEPAGE: HOMEPAGE,
                    CREATED_ON: CREATED_ON,
                    HOMEPAGE_ID: HOMEPAGE_ID,
                    HOMEPAGE_STATUS: HOMEPAGE_STATUS,
                    IS_CONNECTED: IS_CONNECTED
                })
        except Exception:
            print_exc()
            response = []

        homepage_df = pd.DataFrame()
        for query in response:
            df = pd.DataFrame.from_dict(query)
            homepage_df = pd.concat([homepage_df, df], axis=0).reset_index(drop=True)
        homepage_df_active = homepage_df[homepage_df[HOMEPAGE_STATUS] == ACTIVE].reset_index(drop=True)
        return homepage_df_active

    @staticmethod
    def get_json_format_output(df, key_prefix):
        try:
            output_dict = {}
            unique_cluster_id = df[CLUSTER_ID].unique()
            for cluster_id in unique_cluster_id:
                temp_output_dict = {}
                key_prefix_cls = key_prefix + ":" + str(cluster_id)
                output_df = df.loc[df[CLUSTER_ID] == cluster_id]
                output_df = output_df[[HOMEPAGE_ID, CREATED_ON, REC_TYPE]]
                temp_output_dict[key_prefix_cls] = output_df.to_dict(RECORDS)
                output_dict.update(temp_output_dict)
            return output_dict
        except Exception as e:
            Logging.error(f"Error while converting df to json format for{key_prefix}, Error: {e}")

    @staticmethod
    def get_recency_based_results(df):
        try:
            Logging.info("Sorting data based on recency")
            sorted_df = df.sort_values(by=[CLUSTER_ID, CREATED_ON], ascending=(True, False)).reset_index(drop=True)
            return sorted_df
        except Exception as e:
            Logging.error(f"Error while sorting based on recency, Error: {e}")

    @staticmethod
    def get_cluster_wise_homepage_id_not_in_ubd(active_static_homepage_data, cluster_homepage_mapped_user_content_df):
        try:
            Logging.info("Finding cluster-wise homepage_ids not in log")
            active_homepage_df = set(active_static_homepage_data[HOMEPAGE_ID].unique())
            cluster_wise_homepage_id_not_in_ubd = pd.DataFrame()
            unique_cluster_id = cluster_homepage_mapped_user_content_df[CLUSTER_ID].unique()
            for cluster_id in unique_cluster_id:
                ubd = cluster_homepage_mapped_user_content_df.loc[
                    cluster_homepage_mapped_user_content_df[CLUSTER_ID] == cluster_id]
                unique_ubd_homepage = set(ubd[HOMEPAGE_ID].unique())
                homepage_id_not_in_ubd_list = list(active_homepage_df.difference(unique_ubd_homepage))
                homepage_id_not_in_ubd_dict = {CLUSTER_ID: cluster_id, HOMEPAGE_ID: homepage_id_not_in_ubd_list}
                homepage_id_not_in_ubd_df = pd.DataFrame(homepage_id_not_in_ubd_dict)
                cluster_wise_homepage_id_not_in_ubd = pd.concat(
                    [cluster_wise_homepage_id_not_in_ubd, homepage_id_not_in_ubd_df], axis=0).reset_index(drop=True)
            cluster_wise_homepage_id_not_in_ubd = RecencyUtils.get_inner_merged_df(cluster_wise_homepage_id_not_in_ubd,
                                                                                   active_static_homepage_data,
                                                                                   on=HOMEPAGE_ID)
            return cluster_wise_homepage_id_not_in_ubd
        except Exception as e:
            Logging.error(f"Error while Finding cluster-wise homepage_ids not in log, Error: {e}")

    @staticmethod
    def get_default_homepage_id_not_in_ubd(active_static_homepage_data, ubd):
        try:
            Logging.info("Finding homepage_ids not in log")
            active_homepage_df = set(active_static_homepage_data[HOMEPAGE_ID].unique())
            unique_ubd_homepage = set(ubd[HOMEPAGE_ID].unique())
            homepage_id_not_in_ubd_list = list(active_homepage_df.difference(unique_ubd_homepage))
            homepage_id_not_in_ubd_dict = {HOMEPAGE_ID: homepage_id_not_in_ubd_list}
            homepage_id_not_in_ubd_df = pd.DataFrame(homepage_id_not_in_ubd_dict)
            homepage_id_not_in_ubd_df = RecencyUtils.get_inner_merged_df(homepage_id_not_in_ubd_df,
                                                                         active_static_homepage_data,
                                                                         on=HOMEPAGE_ID)
            return homepage_id_not_in_ubd_df
        except Exception as e:
            Logging.error(f"Error while Finding homepage_ids not in log, Error: {e}")

    @staticmethod
    def get_default_recency_based_results(df):
        try:
            Logging.info("Sorting data based on recency")
            sorted_df = df.sort_values(by=[CREATED_ON], ascending=False).reset_index(drop=True)
            return sorted_df
        except Exception as e:
            Logging.error(f"Error while sorting based on recency, Error: {e}")

    @staticmethod
    def get_default_json_format_output(df, key_prefix):
        try:
            output_dict = {}
            key_prefix_cls = key_prefix
            df = df[[HOMEPAGE_ID, CREATED_ON, REC_TYPE]]
            output_dict[key_prefix_cls] = df.to_dict(RECORDS)
            return output_dict
        except Exception as e:
            Logging.error(f"Error while converting df to json format for{key_prefix}, Error: {e}")
