import ast
import time
from operator import itemgetter
from datetime import datetime
import pandas as pd
from sklearn.preprocessing import MinMaxScaler
from functools import reduce

from offline_results.common.config import CONFIG_HOMEPAGE_PAYTV, CONFIG_HOMEPAGE_NO_PAYTV
from offline_results.common.constants import CONTENT_ID, \
    CREATED_ON, VIEW_HISTORY, DURATION, CUSTOMER_ID, INNER, RATING_COUNT, RATING_SUM, \
    RATING_AVERAGE, SCALED_DURATION, SCALED_WEIGHTED_RATING, PAY_TV_CONTENT, \
    NO_PAY_TV_CONTENT, STATUS, HAS_HOMEPAGE, HOMEPAGE_ID, HOMEPAGE_STATUS, ACTIVE, SCORE, \
    SERVICE_NAME, ALL_CONTENT_BASED, POPULAR_MODULE_NAME, RECORDS, HOMEPAGE_ID_BASED, \
    PAY_TV, NO_PAY_TV, IS_PAY_TV, TV_CHANNEL_LIST, RECENT_VIEWED_DATE, IMPLICIT_RATING, \
    WEIGHTED_RATING, POPULAR, REC_TYPE, MODEL_NAME, LEFT
from offline_results.recommendation.utils import RecommendationUtils
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils.logger import Logging


class PopularModel:
    viewed_relation_history_df = pd.DataFrame()
    implicit_rating_df = pd.DataFrame()

    @staticmethod
    def get_data():
        obj = PopularModel
        if len(obj.viewed_relation_history_df) != 0: return
        obj.viewed_relation_history_df = RecommendationUtils.user_viewed_data_from_s3()
        obj.implicit_rating_df = RecommendationUtils.implicit_rating_data_from_s3()

    @staticmethod
    def clear_data():
        obj = PopularModel
        obj.viewed_relation_history_df = pd.DataFrame()
        obj.implicit_rating_df = pd.DataFrame()

    @staticmethod
    def get_recent_viewed_date_and_duration(history):
        history_data = history[VIEW_HISTORY]
        list_history_data = ast.literal_eval(history_data)
        list_history_data = sorted(list_history_data, key=itemgetter(CREATED_ON), reverse=True)
        recent_date = list_history_data[0][CREATED_ON]

        total_duration = 0
        for history in list_history_data:
            duration_content = history[DURATION]
            total_duration += duration_content
        return recent_date, total_duration

    @staticmethod
    def get_json_format_output(df, key_prefix, homepage_id_wise):
        if homepage_id_wise:
            output_dict = {}
            unique_homepage_id = df[HOMEPAGE_ID].unique()
            for homepage_id in unique_homepage_id:
                key_prefix_cls_hid = key_prefix + ":" + str(homepage_id)
                homepage_wise_df1 = df.loc[df[HOMEPAGE_ID] == homepage_id]
                homepage_wise_df1.drop(columns=HOMEPAGE_ID, axis=1, inplace=True)
                output_dict[key_prefix_cls_hid] = homepage_wise_df1.to_dict(RECORDS)
        else:
            df = df[[CONTENT_ID, HOMEPAGE_ID, SCORE, CREATED_ON, REC_TYPE]]
            output_dict = {key_prefix: df.to_dict(RECORDS)}

        return output_dict

    @staticmethod
    def get_content_homepage_id_mapping(graph, user_label, homepage_id_wise):
        content_label = PAY_TV_CONTENT if user_label == PAY_TV else NO_PAY_TV_CONTENT
        queries = graph.custom_query(f'''g.V().hasLabel('{content_label}').match(
                        __.as("c").valueMap("{CONTENT_ID}","{STATUS}").as("{CONTENT_ID}"),
                        __.as("c").out("{HAS_HOMEPAGE}")
                        .valueMap("{HOMEPAGE_ID}","{HOMEPAGE_STATUS}").as("{HOMEPAGE_ID}")
                        ).select("{CONTENT_ID}", "{HOMEPAGE_ID}")''',
                                     payload={content_label: content_label,
                                              CONTENT_ID: CONTENT_ID,
                                              HAS_HOMEPAGE: HAS_HOMEPAGE,
                                              HOMEPAGE_ID: HOMEPAGE_ID,
                                              HOMEPAGE_STATUS: HOMEPAGE_STATUS,
                                              STATUS: STATUS})
        content_homepage_map_df = pd.DataFrame()
        for query in queries:
            for data in query:
                content_id = data[CONTENT_ID][CONTENT_ID][0]
                content_status = data[CONTENT_ID][STATUS][0]
                homepage_id = data[HOMEPAGE_ID][HOMEPAGE_ID][0]
                homepage_status = data[HOMEPAGE_ID][HOMEPAGE_STATUS][0]
                content_homepage_map = pd.DataFrame([{CONTENT_ID: content_id,
                                                      STATUS: content_status,
                                                      HOMEPAGE_ID: homepage_id,
                                                      HOMEPAGE_STATUS: homepage_status}])
                content_homepage_map_df = pd.concat([content_homepage_map_df, content_homepage_map], axis=0) \
                    .sort_values(by=CONTENT_ID).reset_index(drop=True)
        if homepage_id_wise:
            content_homepage_map_df = content_homepage_map_df[
                (content_homepage_map_df[HOMEPAGE_STATUS] == ACTIVE) & (content_homepage_map_df[STATUS] == ACTIVE)
                ].reset_index(drop=True)
        else:
            content_homepage_map_df = content_homepage_map_df[content_homepage_map_df[STATUS] == ACTIVE]\
                .reset_index(drop=True)
        content_homepage_map_df_grouped = pd.DataFrame(
            content_homepage_map_df.groupby(CONTENT_ID)[HOMEPAGE_ID].apply(list).reset_index()
        )
        return content_homepage_map_df_grouped

    @staticmethod
    def filter_by_dedicated_homepage_id(df, user_label, filter_on_ubd):
        config_dictionary = CONFIG_HOMEPAGE_PAYTV if user_label == PAY_TV else CONFIG_HOMEPAGE_NO_PAYTV
        config_df = pd.DataFrame(config_dictionary.items())
        config_df.columns = [HOMEPAGE_ID, MODEL_NAME]
        config_df = config_df[config_df[MODEL_NAME] == POPULAR]
        list_dedicated_homepage_id = config_df[HOMEPAGE_ID].tolist()
        df = df.explode(column=HOMEPAGE_ID)
        df = df[df[HOMEPAGE_ID].isin(list_dedicated_homepage_id)].reset_index(drop=True)
        if filter_on_ubd:
            user_content_network_df = df.drop_duplicates([CONTENT_ID, RECENT_VIEWED_DATE])
        else:
            popular_content_df = df.drop_duplicates([CONTENT_ID, CREATED_ON])
        return user_content_network_df if filter_on_ubd else popular_content_df

    @staticmethod
    def fetch_required_data(graph, user_label, homepage_id_wise):
        try:
            Logging.info("Fetch content-homepage_id mapping")
            content_homepage_id_mapping = PopularModel.get_content_homepage_id_mapping(graph, user_label,
                                                                                       homepage_id_wise)
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Re-trying to Fetch content-homepage_id mapping")
            content_homepage_id_mapping = PopularModel.get_content_homepage_id_mapping(graph, user_label,
                                                                                       homepage_id_wise)
        Logging.info("Load Implicit Rating Network from S3")
        is_pay_tv_status = True if user_label == PAY_TV else False
        implicit_rating_df = PopularModel.implicit_rating_df
        implicit_rating_df = implicit_rating_df[implicit_rating_df[IS_PAY_TV] == is_pay_tv_status]
        implicit_rating_df = implicit_rating_df[~implicit_rating_df[CONTENT_ID].isin(TV_CHANNEL_LIST)]\
            .reset_index(drop=True)
        Logging.info("Load Viewed Relation Network from S3")
        viewed_relation_history_df = PopularModel.viewed_relation_history_df
        viewed_relation_history_df = viewed_relation_history_df[viewed_relation_history_df[IS_PAY_TV] == is_pay_tv_status]
        viewed_relation_history_df = viewed_relation_history_df[~viewed_relation_history_df[CONTENT_ID].isin(TV_CHANNEL_LIST)].reset_index(drop=True)
        Logging.info('Generating Viewed Relation Dataframe for ' + user_label + ' users')
        view_history = viewed_relation_history_df[VIEW_HISTORY].to_list()
        view_history2 = [ast.literal_eval(i) for i in view_history]
        intermediate_df = pd.DataFrame({CUSTOMER_ID: viewed_relation_history_df[CUSTOMER_ID].to_list(),
                                        CONTENT_ID: viewed_relation_history_df[CONTENT_ID].to_list(),
                                        VIEW_HISTORY: view_history2})
        Logging.info('Get user cluster mapping')
        intermediate_df = intermediate_df.explode(column=VIEW_HISTORY)
        intermediate_df[DURATION] = intermediate_df[VIEW_HISTORY].apply(lambda x: x[DURATION])
        intermediate_df[RECENT_VIEWED_DATE] = intermediate_df[VIEW_HISTORY].apply(lambda x: x[CREATED_ON])
        recent_date_df = intermediate_df.groupby(CONTENT_ID).agg({RECENT_VIEWED_DATE: 'max'}).rename_axis(CONTENT_ID) \
            .reset_index()
        total_duration_df = intermediate_df.groupby(CONTENT_ID).sum().rename_axis(CONTENT_ID).reset_index()[[CONTENT_ID,
                                                                                                             DURATION]]
        data_frames = [recent_date_df, total_duration_df, content_homepage_id_mapping]
        user_content_network_df = reduce(lambda left, right: pd.merge(left, right, on=[CONTENT_ID], how=LEFT),
                                         data_frames)
        user_content_network_df = user_content_network_df[~user_content_network_df[HOMEPAGE_ID].isnull()].reset_index(drop=True)
        user_content_network_df = PopularModel.filter_by_dedicated_homepage_id(user_content_network_df, user_label,
                                                                               filter_on_ubd=True)
        user_content_network_df = pd.merge(user_content_network_df, implicit_rating_df, on=[CONTENT_ID], how=LEFT)
        user_content_network_df = user_content_network_df[~user_content_network_df[IMPLICIT_RATING].isnull()].reset_index(drop=True)
        user_content_network_df = user_content_network_df.drop_duplicates([CONTENT_ID, CUSTOMER_ID, RECENT_VIEWED_DATE]).reset_index(drop=True)
        return user_content_network_df, content_homepage_id_mapping

    @staticmethod
    def preprocessed_required_data(data):
        Logging.info("Preprocessing Required Data")
        rating_count_df = data.groupby([CONTENT_ID]).size().to_frame(RATING_COUNT)
        rating_sum_df = data.groupby(CONTENT_ID).sum().rename(columns={IMPLICIT_RATING: RATING_SUM})
        prepared_df = rating_sum_df.merge(rating_count_df, how=INNER, left_index=True, right_index=True)
        prepared_df[RATING_AVERAGE] = prepared_df[RATING_SUM] / prepared_df[RATING_COUNT]

        return prepared_df

    @staticmethod
    def popular_algorithm(graph, user_label, homepage_id_wise):
        required_data, content_homepage_id_mapping = PopularModel.fetch_required_data(graph, user_label,
                                                                                      homepage_id_wise)
        prepared_data = PopularModel.preprocessed_required_data(required_data)

        Logging.info("Calculating Weighted Rating")
        C = prepared_data[RATING_AVERAGE].mean()
        m = prepared_data[RATING_COUNT].quantile(0.3)
        listed_content_df = prepared_data.loc[prepared_data[RATING_COUNT] > m]

        def weighted_rating_formula(prepared_data, m=m, C=C):
            v = prepared_data[RATING_COUNT]
            R = prepared_data[RATING_AVERAGE]

            return (v / (v + m) * R) + (m / (m + v) * C)

        listed_content_df[WEIGHTED_RATING] = listed_content_df.apply(
            weighted_rating_formula, axis=1
        )
        listed_content_df = listed_content_df.sort_values(by=WEIGHTED_RATING, ascending=False)

        selected_df = listed_content_df[[DURATION, WEIGHTED_RATING]]
        scaled_features_df = MinMaxScaler().fit_transform(selected_df.values)
        scaled_features_df = pd.DataFrame(scaled_features_df, index=selected_df.index, columns=selected_df.columns) \
            .rename(columns={DURATION: SCALED_DURATION, WEIGHTED_RATING: SCALED_WEIGHTED_RATING})
        listed_content_df = listed_content_df.merge(scaled_features_df, how=INNER, left_index=True, right_index=True)
        Logging.info("Calculating Popular Score from " + user_label + " UBD data")
        listed_content_df[SCORE] = round((listed_content_df[SCALED_DURATION] * 0.5) +
                                         (listed_content_df[SCALED_WEIGHTED_RATING] * 0.5), 3)
        listed_content_df = listed_content_df.sort_values(by=SCORE, ascending=False)
        Logging.info("Preparing Popular Model Dataframe Output")
        listed_content_df[CREATED_ON] = datetime.utcnow().isoformat()
        listed_content_df[REC_TYPE] = POPULAR
        popular_content_df = listed_content_df[[SCORE, CREATED_ON, REC_TYPE]].rename_axis(CONTENT_ID).reset_index()

        popular_content_df = pd.merge(popular_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=INNER)
        popular_content_df = popular_content_df[~popular_content_df[HOMEPAGE_ID].isnull()].reset_index(drop=True)
        popular_content_df = popular_content_df.explode(column=HOMEPAGE_ID)
        popular_content_df = PopularModel.filter_by_dedicated_homepage_id(popular_content_df, user_label,
                                                                          filter_on_ubd=False)
        if homepage_id_wise:
            popular_content_df = popular_content_df.sort_values(by=[HOMEPAGE_ID, SCORE], ascending=[True, False])
        else:
            popular_content_df = popular_content_df.sort_values(by=SCORE, ascending=False)

        return popular_content_df

    @staticmethod
    def get_popular_content_all_content(graph):
        PopularModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = False
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Popular Content based on All Content")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating popular content based on All Content for " + user_label + " users")
            output_df = PopularModel.popular_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Popular Dataframe Output")
            key_prefix = SERVICE_NAME + ":" + POPULAR_MODULE_NAME + ":" + user_label + ":" + ALL_CONTENT_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        popular_pay_tv_all_content, popular_no_pay_tv_all_content = list_output_df
        Logging.info("Preparing the output to JSON Schema")
        popular_pay_tv_all_content_json = PopularModel.get_json_format_output(popular_pay_tv_all_content,
                                                                              list_of_key_prefix[0], homepage_id_wise)
        popular_no_pay_tv_all_content_json = PopularModel.get_json_format_output(popular_no_pay_tv_all_content,
                                                                                 list_of_key_prefix[1],
                                                                                 homepage_id_wise)
        end = time.time()
        duration = end - start
        Logging.info("Duration for Get Popular Content based on All Content is " + str(duration) + " seconds")
        PopularModel.clear_data()
        return popular_pay_tv_all_content_json, popular_no_pay_tv_all_content_json

    @staticmethod
    def get_popular_content_homepage_id(graph):
        PopularModel.clear_data()
        PopularModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = True
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Popular Content based on Homepage_id")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating popular content based on Homepage_id for " + user_label + " users")
            output_df = PopularModel.popular_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Popular Dataframe Output")
            key_prefix = SERVICE_NAME + ":" + POPULAR_MODULE_NAME + ":" + user_label + ":" + HOMEPAGE_ID_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        popular_pay_tv_homepage_id, popular_no_pay_tv_homepage_id = list_output_df
        Logging.info("Preparing the output to JSON Schema")
        popular_pay_tv_homepage_id_json = PopularModel.get_json_format_output(popular_pay_tv_homepage_id,
                                                                              list_of_key_prefix[0], homepage_id_wise)
        popular_no_pay_tv_homepage_id_json = PopularModel.get_json_format_output(popular_no_pay_tv_homepage_id,
                                                                                 list_of_key_prefix[1],
                                                                                 homepage_id_wise)
        end = time.time()
        duration = end - start
        Logging.info("Duration for Get Popular Content based on Homepage_id is " + str(duration) + " seconds")
        return popular_pay_tv_homepage_id_json, popular_no_pay_tv_homepage_id_json
