import time
from datetime import datetime
import ast
from operator import itemgetter
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

from offline_results.common.config import CONFIG_HOMEPAGE_PAYTV, CONFIG_HOMEPAGE_NO_PAYTV
from offline_results.common.constants import CONTENT_ID, CREATED_ON, VIEW_HISTORY, VIEW_COUNT, UBD_WEEKOFYEAR, \
    DURATION, UBD_CREATED_ON, UBD_DAYOFWEEK, SCALED_DURATION, SCALED_VIEW_COUNT, INNER, PAY_TV_CONTENT, \
    NO_PAY_TV_CONTENT, HAS_HOMEPAGE, HOMEPAGE_ID, SCORE, HOMEPAGE_STATUS, STATUS, ACTIVE, RECORDS, \
    SERVICE_NAME, TRENDING_MODULE_NAME, ALL_CONTENT_BASED, HOMEPAGE_ID_BASED, \
    TV_CHANNEL_LIST, PAY_TV, NO_PAY_TV, IS_PAY_TV, REC_TYPE, TRENDING, LEFT, MODEL_NAME
from offline_results.recommendation.utils import RecommendationUtils
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils.logger import Logging


class TrendingModel:

    viewed_relation_history_df = pd.DataFrame()

    @staticmethod
    def get_data():
        obj = TrendingModel
        if len(obj.viewed_relation_history_df) != 0: return
        obj.viewed_relation_history_df = RecommendationUtils.user_viewed_data_from_s3()

    @staticmethod
    def clear_data():
        obj = TrendingModel
        obj.viewed_relation_history_df = pd.DataFrame()

    @staticmethod
    def filter_by_dedicated_homepage_id(df, user_label, filter_on_ubd):
        config_dictionary = CONFIG_HOMEPAGE_PAYTV if user_label == PAY_TV else CONFIG_HOMEPAGE_NO_PAYTV
        config_df = pd.DataFrame(config_dictionary.items())
        config_df.columns = [HOMEPAGE_ID, MODEL_NAME]
        config_df = config_df[config_df[MODEL_NAME] == TRENDING]
        list_dedicated_homepage_id = config_df[HOMEPAGE_ID].tolist()
        df = df.explode(column=HOMEPAGE_ID)
        df = df[df[HOMEPAGE_ID].isin(list_dedicated_homepage_id)].reset_index(drop=True)
        if filter_on_ubd:
            history_content_df = df.drop_duplicates([CONTENT_ID, UBD_CREATED_ON])
        else:
            trending_content_df = df.drop_duplicates([CONTENT_ID, CREATED_ON])
        return history_content_df if filter_on_ubd else trending_content_df

    @staticmethod
    def get_dict_format_output(df, key_prefix, homepage_id_wise):
        if homepage_id_wise:
            output_dict = {}
            unique_homepage_id = df[HOMEPAGE_ID].unique()
            for homepage_id in unique_homepage_id:
                key_prefix_cls_hid = key_prefix + ":" + str(homepage_id)
                homepage_wise_df = df.loc[df[HOMEPAGE_ID] == homepage_id]
                homepage_wise_df = homepage_wise_df[[CONTENT_ID, SCORE, CREATED_ON, REC_TYPE]]
                output_dict[key_prefix_cls_hid] = homepage_wise_df.to_dict(RECORDS)
        else:
            df = df[[CONTENT_ID, HOMEPAGE_ID, SCORE, CREATED_ON, REC_TYPE]]
            output_dict = {key_prefix: df.to_dict(RECORDS)}
        return output_dict

    @staticmethod
    def get_history_date(history):
        history_data = history[VIEW_HISTORY]
        list_history_data = ast.literal_eval(history_data)
        list_history_data = sorted(list_history_data, key=itemgetter(CREATED_ON), reverse=True)
        history_date = pd.DataFrame(list_history_data).rename(columns={CREATED_ON: UBD_CREATED_ON})
        return history_date

    @staticmethod
    def get_content_homepage_id_mapping(graph, user_label, homepage_id_wise):
        content_label = PAY_TV_CONTENT if user_label == PAY_TV else NO_PAY_TV_CONTENT
        queries = graph.custom_query(f'''g.V().hasLabel('{content_label}').match(
                        __.as("c").valueMap("{CONTENT_ID}","{STATUS}").as("{CONTENT_ID}"),
                        __.as("c").out("{HAS_HOMEPAGE}")
                        .valueMap("{HOMEPAGE_ID}","{HOMEPAGE_STATUS}").as("{HOMEPAGE_ID}")
                        ).select("{CONTENT_ID}", "{HOMEPAGE_ID}")''',
                                     payload={content_label: content_label,
                                              CONTENT_ID: CONTENT_ID,
                                              HAS_HOMEPAGE: HAS_HOMEPAGE,
                                              HOMEPAGE_ID: HOMEPAGE_ID,
                                              HOMEPAGE_STATUS: HOMEPAGE_STATUS,
                                              STATUS: STATUS})
        content_homepage_map_df = pd.DataFrame()
        for query in queries:
            for data in query:
                content_id = data[CONTENT_ID][CONTENT_ID][0]
                content_status = data[CONTENT_ID][STATUS][0]
                homepage_id = data[HOMEPAGE_ID][HOMEPAGE_ID][0]
                homepage_status = data[HOMEPAGE_ID][HOMEPAGE_STATUS][0]
                content_homepage_map = pd.DataFrame([{CONTENT_ID: content_id,
                                                      STATUS: content_status,
                                                      HOMEPAGE_ID: homepage_id,
                                                      HOMEPAGE_STATUS: homepage_status}])
                content_homepage_map_df = pd.concat([content_homepage_map_df, content_homepage_map], axis=0)\
                    .sort_values(by=CONTENT_ID).reset_index(drop=True)
        if homepage_id_wise:
            content_homepage_map_df = content_homepage_map_df[
                (content_homepage_map_df[HOMEPAGE_STATUS] == ACTIVE) & (content_homepage_map_df[STATUS] == ACTIVE)].reset_index(drop=True)
        else:
            content_homepage_map_df = content_homepage_map_df[content_homepage_map_df[STATUS] == ACTIVE].reset_index(drop=True)
        content_homepage_map_df_grouped = pd.DataFrame(
            content_homepage_map_df.groupby(CONTENT_ID)[HOMEPAGE_ID].apply(list).reset_index()
        )
        return content_homepage_map_df_grouped

    @staticmethod
    def fetch_historical_data(graph, user_label, homepage_id_wise):
        try:
            Logging.info("Fetch content-homepage_id mapping")
            content_homepage_id_mapping = TrendingModel.get_content_homepage_id_mapping(graph, user_label,
                                                                                        homepage_id_wise)
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Re-trying to Fetch content-homepage_id mapping")
            content_homepage_id_mapping = TrendingModel.get_content_homepage_id_mapping(graph, user_label,
                                                                                        homepage_id_wise)
        Logging.info("Load Viewed Relation Network from S3")
        viewed_relation_history_df = TrendingModel.viewed_relation_history_df
        is_pay_tv_status = True if user_label == PAY_TV else False
        viewed_relation_history_df = viewed_relation_history_df[viewed_relation_history_df[IS_PAY_TV] == is_pay_tv_status]
        viewed_relation_history_df = viewed_relation_history_df[~viewed_relation_history_df[CONTENT_ID].isin(TV_CHANNEL_LIST)].reset_index(drop=True)
        view_history = viewed_relation_history_df[VIEW_HISTORY].to_list()
        view_history2 = [ast.literal_eval(i) for i in view_history]
        history_content_df = pd.DataFrame({CONTENT_ID: viewed_relation_history_df[CONTENT_ID].to_list(),
                                           VIEW_HISTORY: view_history2})
        history_content_df = history_content_df.explode(column=VIEW_HISTORY)
        history_content_df[DURATION] = history_content_df[VIEW_HISTORY].apply(lambda x: x[DURATION])
        history_content_df[UBD_CREATED_ON] = history_content_df[VIEW_HISTORY].apply(lambda x: x[CREATED_ON])
        history_content_df[UBD_CREATED_ON] = pd.to_datetime(history_content_df[UBD_CREATED_ON], dayfirst=True)
        history_content_df = history_content_df.sort_values(by=UBD_CREATED_ON, ascending=False).reset_index(drop=True)
        history_content_df[UBD_WEEKOFYEAR] = history_content_df[UBD_CREATED_ON].dt.isocalendar().week
        history_content_df[UBD_DAYOFWEEK] = history_content_df[UBD_CREATED_ON].dt.dayofweek
        history_content_df = pd.merge(history_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=LEFT)
        Logging.info("Filter by dedicated homepage_id for Trending Model")
        history_content_df = TrendingModel.filter_by_dedicated_homepage_id(history_content_df, user_label,
                                                                           filter_on_ubd=True)
        return history_content_df, content_homepage_id_mapping

    @staticmethod
    def get_latest_week_df(history_content_df):
        Logging.info("Get last week UBD Dataframe data")
        get_latest_week_number = history_content_df[UBD_WEEKOFYEAR].unique()[0]
        latest_week_df = history_content_df[history_content_df[UBD_WEEKOFYEAR] == get_latest_week_number]
        total_dayofweek = len(latest_week_df[UBD_DAYOFWEEK].unique())
        if total_dayofweek < 4:
            two_latest_week_number = history_content_df[UBD_WEEKOFYEAR].unique()[0:2].tolist()
            latest_week_df = history_content_df[history_content_df[UBD_WEEKOFYEAR].isin(two_latest_week_number)]
            return latest_week_df

        return latest_week_df

    @staticmethod
    def trending_algorithm(graph, user_label, homepage_id_wise):
        history_content_df, content_homepage_id_mapping = TrendingModel.fetch_historical_data(graph, user_label,
                                                                                              homepage_id_wise)
        latest_week_df = TrendingModel.get_latest_week_df(history_content_df)

        Logging.info("Calculating Trending Score from " + user_label + " UBD data")
        view_count_on_latest_week_df = latest_week_df.groupby([CONTENT_ID]).size().to_frame(VIEW_COUNT)
        total_duration_content = latest_week_df.groupby([CONTENT_ID]).sum()
        total_duration_content = total_duration_content[[DURATION]]
        final_recommendation_df = pd.merge(total_duration_content, view_count_on_latest_week_df,
                                           left_index=True, right_index=True).sort_values(by=[DURATION, VIEW_COUNT],
                                                                                          ascending=False)
        scaled_features_df = MinMaxScaler().fit_transform(final_recommendation_df.values)
        scaled_features_df = pd.DataFrame(scaled_features_df, index=final_recommendation_df.index,
                                          columns=final_recommendation_df.columns).rename(
            columns={DURATION: SCALED_DURATION, VIEW_COUNT: SCALED_VIEW_COUNT}
        )
        final_recommendation_df = final_recommendation_df.merge(scaled_features_df, how=INNER,
                                                                left_index=True, right_index=True)
        final_recommendation_df[SCORE] = round((final_recommendation_df[SCALED_DURATION]*0.7) +
                                               (final_recommendation_df[SCALED_VIEW_COUNT]*0.3), 3)
        final_recommendation_df = final_recommendation_df.sort_values(by=SCORE, ascending=False)

        Logging.info("Preparing Trending Model Dataframe Output")
        final_recommendation_df[CREATED_ON] = datetime.utcnow().isoformat()
        trending_content_df = final_recommendation_df[[SCORE, CREATED_ON]].rename_axis(CONTENT_ID).reset_index()
        trending_content_df[REC_TYPE] = TRENDING

        trending_content_df = pd.merge(trending_content_df, content_homepage_id_mapping, on=CONTENT_ID, how=INNER)
        trending_content_df = trending_content_df[~trending_content_df[HOMEPAGE_ID].isnull()].reset_index(drop=True)
        trending_content_df = trending_content_df.explode(column=HOMEPAGE_ID)
        trending_content_df = TrendingModel.filter_by_dedicated_homepage_id(trending_content_df, user_label,
                                                                          filter_on_ubd=False)
        if homepage_id_wise:
            trending_content_df = trending_content_df.sort_values(by=[HOMEPAGE_ID, SCORE], ascending=[True, False])
        else:
            trending_content_df = trending_content_df.sort_values(by=SCORE, ascending=False)
        return trending_content_df

    @staticmethod
    def get_trending_content_homepage_id(graph):
        TrendingModel.clear_data()
        TrendingModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = True
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Trending Content based on Homepage_id")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating trending content based on Homepage_id for " + user_label + " users")
            output_df = TrendingModel.trending_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Trending Dataframe Output")
            key_prefix = SERVICE_NAME + ":" + TRENDING_MODULE_NAME + ":" + user_label + ":" + HOMEPAGE_ID_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        trending_pay_tv_homepage_id, trending_no_pay_tv_homepage_id = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        trending_pay_tv_homepage_id_dict = TrendingModel.get_dict_format_output(trending_pay_tv_homepage_id,
                                                                                list_of_key_prefix[0],
                                                                                homepage_id_wise)
        trending_no_pay_tv_homepage_id_dict = TrendingModel.get_dict_format_output(trending_no_pay_tv_homepage_id,
                                                                                   list_of_key_prefix[1],
                                                                                   homepage_id_wise)
        end = time.time()
        duration = end-start
        Logging.info("Duration for Get Trending Content based on Homepage_id is " + str(duration) + " seconds")
        return trending_pay_tv_homepage_id_dict, trending_no_pay_tv_homepage_id_dict

    @staticmethod
    def get_trending_content_all_content(graph):
        TrendingModel.get_data()
        list_of_key_prefix = []
        homepage_id_wise = False
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_df = []
        Logging.info("Get Trending Content based on All Content")
        start = time.time()
        for user_label in list_user_label:
            Logging.info("Start generating trending content based on All Content for " + user_label + " users")
            output_df = TrendingModel.trending_algorithm(graph, user_label, homepage_id_wise)
            Logging.info("Success Preparing Trending Dataframe Output")
            key_prefix = SERVICE_NAME + ":" + TRENDING_MODULE_NAME + ":" + user_label + ":" + ALL_CONTENT_BASED
            list_of_key_prefix.append(key_prefix)
            list_output_df.append(output_df)
        trending_pay_tv_all_content, trending_no_pay_tv_all_content = list_output_df

        Logging.info("Preparing the output to JSON Schema")
        # Convert dataframe output to json output
        trending_pay_tv_all_content_dict = TrendingModel.get_dict_format_output(trending_pay_tv_all_content,
                                                                                list_of_key_prefix[0],
                                                                                homepage_id_wise)
        trending_no_pay_tv_all_content_dict = TrendingModel.get_dict_format_output(trending_no_pay_tv_all_content,
                                                                                   list_of_key_prefix[1],
                                                                                   homepage_id_wise)
        end = time.time()
        duration = end-start
        Logging.info("Duration for Get Trending Content based on All Content is " + str(duration) + " seconds")
        TrendingModel.clear_data()
        return trending_pay_tv_all_content_dict, trending_no_pay_tv_all_content_dict
